#!/usr/bin/env python
"""

Auxiliary functions for f2py2e.

Copyright 1999,2000 Pearu Peterson all rights reserved,
Pearu Peterson <pearu@ioc.ee>          
Permission to use, modify, and distribute this software is given under the
terms of the LGPL.  See http://www.fsf.org

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
$Date: 2002/03/05 14:21:27 $
Pearu Peterson
"""
__version__ = "$Revision: 1.34 $"[10:-1]

import __version__
f2py_version = __version__.version

import pprint
import sys,string,time,types,os
import cfuncs


errmess=sys.stderr.write
outmess=sys.stdout.write
show=pprint.pprint

debugoptions=[]
wrapfuncs = 1
def debugcapi(var): return 'capi' in debugoptions
def isstring(var):
    return var.has_key('typespec') and var['typespec']=='character' and (not isexternal(var))
def isstringarray(var): return isarray(var) and isstring(var)
def isarray(var): return var.has_key('dimension') and (not isexternal(var))
def isscalar(var): return not (isarray(var) or isstring(var) or isexternal(var))
def iscomplex(var):
    return isscalar(var) and var['typespec'] in ['complex','double complex']
def islong_long(var):
    if not (var.has_key('kindselector') and var['kindselector'].has_key('*')): return 0
    return isscalar(var) and var['typespec']=='integer' and var['kindselector']['*']=='8'
def islong_double(var):
    if not (var.has_key('kindselector') and var['kindselector'].has_key('*')): return 0
    return isscalar(var) and var['typespec']=='real' and var['kindselector']['*']=='16'
def islong_complex(var):
    if not (var.has_key('kindselector') and var['kindselector'].has_key('*')): return 0
    return iscomplex(var) and var['kindselector']['*']=='32'
def iscomplexarray(var): return isarray(var) and var['typespec'] in ['complex','double complex']
def isallocatable(var):
    return var.has_key('attrspec') and 'allocatable' in var['attrspec']
def ismutable(var): return not (not var.has_key('dimension') or isstring(var))
def ismoduleroutine(rout): return rout.has_key('modulename')
def ismodule(rout): return (rout.has_key('block') and 'module'==rout['block'])
def isfunction(rout): return (rout.has_key('block') and 'function'==rout['block'])
#def isfunction_wrap(rout):
#    return wrapfuncs and (iscomplexfunction(rout) or isstringfunction(rout)) and (not isexternal(rout))
def isfunction_wrap(rout):
    if isintent_c(rout): return 0
    return wrapfuncs and isfunction(rout) and (not isexternal(rout))
def issubroutine(rout): return (rout.has_key('block') and 'subroutine'==rout['block'])
def isroutine(rout): return isfunction(rout) or issubroutine(rout)
def islong_longfunction(rout):
    if not isfunction(rout): return 0
    if rout.has_key('result'): a=rout['result']
    else: a=rout['name']
    if rout['vars'].has_key(a): return islong_long(rout['vars'][a])
    return 0
def islong_doublefunction(rout):
    if not isfunction(rout): return 0
    if rout.has_key('result'): a=rout['result']
    else: a=rout['name']
    if rout['vars'].has_key(a): return islong_double(rout['vars'][a])
    return 0
def iscomplexfunction(rout):
    if not isfunction(rout): return 0
    if rout.has_key('result'): a=rout['result']
    else: a=rout['name']
    if rout['vars'].has_key(a): return iscomplex(rout['vars'][a])
    return 0
def iscomplexfunction_warn(rout):
    if iscomplexfunction(rout):
        outmess("""\
    **************************************************************
        Warning: code with a function returning complex value
        may not work correctly with your Fortran compiler.
        Run the following test before using it in your applications:
        $(f2py install dir)/test-site/{b/runme_scalar,e/runme}
        When using GNU gcc/g77 compilers, codes should work correctly.
    **************************************************************\n""")
        return 1
    return 0
def isstringfunction(rout):
    if not isfunction(rout): return 0
    if rout.has_key('result'): a=rout['result']
    else: a=rout['name']
    if rout['vars'].has_key(a): return isstring(rout['vars'][a])
    return 0
def hasexternals(rout): return rout.has_key('externals') and rout['externals']
def isthreadsafe(rout): return rout.has_key('f2pyenhancements') and rout['f2pyenhancements'].has_key('threadsafe')
def hasvariables(rout): return rout.has_key('vars') and rout['vars']
def isoptional(var): return (var.has_key('attrspec') and 'optional' in var['attrspec'] and 'required' not in var['attrspec'])
def isexternal(var): return (var.has_key('attrspec') and 'external' in var['attrspec'])
def isrequired(var): return not isoptional(var)
def isintent_in(var): return (var.has_key('intent') and 'in' in var['intent'] and 'hide' not in var['intent']) or (not var.has_key('intent'))
def isintent_inout(var): return var.has_key('intent') and ('inout' in var['intent'] or 'outin' in var['intent']) and 'in' not in var['intent'] and 'hide' not in var['intent']
def isintent_out(var): return (var.has_key('intent') and 'out' in var['intent'])
def isintent_hide(var): return (var.has_key('intent') and ('hide' in var['intent'] or ('out' in var['intent'] and 'in' not in var['intent'] and (not isintent_inout(var)))))
def isintent_nothide(var): return not isintent_hide(var)
def isintent_c(var):
    return (var.has_key('intent') and 'c' in var['intent'])
# def isintent_f(var):
#     return not isintent_c(var)
def isintent_cache(var):
    return (var.has_key('intent') and 'cache' in var['intent'])
def isintent_copy(var):
    return (var.has_key('intent') and 'copy' in var['intent'])
def isintent_overwrite(var):
    return (var.has_key('intent') and 'overwrite' in var['intent'])

isintent_dict = {isintent_in:'INTENT_IN',isintent_inout:'INTENT_INOUT',
                 isintent_out:'INTENT_OUT',isintent_hide:'INTENT_HIDE',
                 isintent_cache:'INTENT_CACHE',
                 isintent_c:'INTENT_C',isoptional:'OPTIONAL',
                 }


def hasinitvalue(var): return var.has_key('=')
def hasinitvalueasstring(var):
    if not hasinitvalue(var): return 0
    return var['='][0] in ['"',"'"]
def hasnote(var):
    return var.has_key('note')
def hasresultnote(rout):
    if not isfunction(rout): return 0
    if rout.has_key('result'): a=rout['result']
    else: a=rout['name']
    if rout['vars'].has_key(a): return hasnote(rout['vars'][a])
    return 0
def hascommon(rout):
    return rout.has_key('common')
def containscommon(rout):
    if hascommon(rout): return 1
    if hasbody(rout):
        for b in rout['body']:
            if containscommon(b): return 1
    return 0
def containsmodule(block):
    if ismodule(block): return 1
    if not hasbody(block): return 0
    ret = []
    for b in block['body']:
        if containsmodule(b): return 1
    return 0
def hasbody(rout):
    return rout.has_key('body')
def hascallstatement(rout):
    return getcallstatement(rout) is not None

def istrue(var): return 1
def isfalse(var): return 0

def l_and(*f):
    l,l2='lambda v',[]
    for i in range(len(f)):
        l='%s,f%d=f[%d]'%(l,i,i)
        l2.append('f%d(v)'%(i))
    return eval('%s:%s'%(l,string.join(l2,' and ')))
def l_or(*f):
    l,l2='lambda v',[]
    for i in range(len(f)):
        l='%s,f%d=f[%d]'%(l,i,i)
        l2.append('f%d(v)'%(i))
    return eval('%s:%s'%(l,string.join(l2,' or ')))
def l_not(f):
    return eval('lambda v,f=f:not f(v)')
def getfortranname(rout):
    try:
        name = rout['f2pyenhancements']['fortranname']
        if not name:
            errmess('Failed to use fortranname from %s'%(rout['f2pyenhancements']))
            raise KeyError
    except KeyError:
        name = rout['name']
    return name
def getcallstatement(rout):
    try:
        return rout['f2pyenhancements'].get('callstatement')
    except KeyError:
        pass
def getcallprotoargument(rout):
    try:
        return rout['f2pyenhancements'].get('callprotoargument')
    except KeyError:
        pass
def getargs(rout):
    sortargs,args=[],[]
    if rout.has_key('args'):
        args=rout['args']
        if rout.has_key('sortvars'):
            for a in rout['sortvars']:
                if a in args: sortargs.append(a)
            for a in args:
                if a not in sortargs:
                    sortargs.append(a)
        else: sortargs=rout['args']
    return args,sortargs

def gentitle(name):
    l=(80-len(name)-6)/2
    return '/*%s %s %s*/'%(l*'*',name,l*'*')
def flatlist(l):
    if type(l)==types.ListType:
        return reduce(lambda x,y,f=flatlist:x+f(y),l,[])
    return [l]
def stripcomma(s):
    if s and s[-1]==',': return s[:-1]
    return s
def replace(str,dict,defaultsep=''):
    if type(dict)==types.ListType:
        return map(lambda d,f=replace,sep=defaultsep,s=str:f(s,d,sep),dict)
    for k in 2*dict.keys():
        if k=='separatorsfor': continue
        if dict.has_key('separatorsfor') and dict['separatorsfor'].has_key(k):
            sep=dict['separatorsfor'][k]
        else:
            sep=defaultsep
        if type(dict[k])==types.ListType:
            str=string.replace(str,'#%s#'%(k),string.join(flatlist(dict[k]),sep))
        else:
            str=string.replace(str,'#%s#'%(k),dict[k])
    return str

def dictappend(rd,ar):
    if type(ar)==types.ListType:
        for a in ar: rd=dictappend(rd,a)
        return rd
    for k in ar.keys():
        if k[0]=='_': continue
        if rd.has_key(k):
            if type(rd[k])==types.StringType: rd[k]=[rd[k]]
            if type(rd[k])==types.ListType:
                if type(ar[k])==types.ListType: rd[k]=rd[k]+ar[k]
                else: rd[k].append(ar[k])
            elif type(rd[k])==types.DictType:
                if type(ar[k])==types.DictType:
                    if k=='separatorsfor':
                        for k1 in ar[k].keys():
                            if not rd[k].has_key(k1): rd[k][k1]=ar[k][k1]
                    else: rd[k]=dictappend(rd[k],ar[k])
        else: rd[k]=ar[k]
    return rd

def applyrules(rules,dict,var={}):
    ret={}
    if type(rules)==types.ListType:
        for r in rules:
            rr=applyrules(r,dict,var)
            ret=dictappend(ret,rr)
            if rr.has_key('_break'): break
        return ret
    if rules.has_key('_check') and (not rules['_check'](var)): return ret
    if rules.has_key('need'):
        res = applyrules({'needs':rules['need']},dict,var)
        if res.has_key('needs'):
            cfuncs.append_needs(res['needs'])

    for k in rules.keys():
        if k=='separatorsfor': ret[k]=rules[k]; continue
        if type(rules[k])==types.StringType:
            ret[k]=replace(rules[k],dict)
        elif type(rules[k])==types.ListType:
            ret[k]=[]
            for i in rules[k]:
                ar=applyrules({k:i},dict,var)
                if ar.has_key(k): ret[k].append(ar[k])
        elif k[0]=='_':
            continue
        elif type(rules[k])==types.DictType:
            ret[k]=[]
            for k1 in rules[k].keys():
                if type(k1)==types.FunctionType and k1(var):
                    if type(rules[k][k1])==types.ListType:
                        for i in rules[k][k1]:
                            if type(i)==types.DictType:
                                res=applyrules({'supertext':i},dict,var)
                                if res.has_key('supertext'): i=res['supertext']
                                else: i=''
                            ret[k].append(replace(i,dict))
                    else:
                        i=rules[k][k1]
                        if type(i)==types.DictType:
                            res=applyrules({'supertext':i},dict)
                            if res.has_key('supertext'): i=res['supertext']
                            else: i=''
                        ret[k].append(replace(i,dict))
        else:
            errmess('applyrules: ignoring rule %s.\n'%`rules[k]`)
        if type(ret[k])==types.ListType:
            if len(ret[k])==1: ret[k]=ret[k][0]
            if ret[k]==[]: del ret[k]
    return ret


