#!/usr/bin/env python
"""

buildmakefile --- builds Makefile-<modulename> (see function build() below).
    Below are defined some Fortran and C compilers, and linkers.
    Add your own (and send me too).

Copyright 2000 Pearu Peterson all rights reserved,
Pearu Peterson <pearu@ioc.ee>          
Permission to use, modify, and distribute this software is given under the
terms of the LGPL.  See http://www.fsf.org

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
$Date: 2002/01/11 16:34:16 $
Pearu Peterson
"""

__version__ = "$Revision: 1.54 $"[10:-1]

import __version__
f2py_version = __version__.version


import types,string,sys,os,time,re,commands,pprint

from auxfuncs import flatlist,dictappend,isfunction_wrap,isfunction,\
     issubroutine,l_or,containscommon,containsmodule
outmess=sys.stdout.write

show = pprint.pprint
fortran_compilers={}
c_compilers={}
linkers={}

# Intel(R) Fortran Compiler (thanks to Patrick LeGresley)
fortran_compilers['Intel(R) Fortran Compiler'] = {
        'versmatch' : r'Intel\(R\) Fortran Compiler for 32-bit applications, Version (?P<version>[^\s]*)',
        'FFLAGS' : '',
        'FOPT'   : '-O3',
        'FDEBUG' : '-g -C',
        'NOTE' : 'use LD=ifc when linking',
        'COMPILERULES' : '%.o : %.f90 ; $(FC) -c $(FFLAGS) $< -o $@'
        }
linkers['Intel(R) Fortran Compiler'] = {
     'versmatch' : r'Intel\(R\) Fortran Compiler for 32-bit applications, Version (?P<version>[^\s]*)',
     'LDFLAGS' : '-shared'
     }


# NAGWare Fortran 95 compiler on linux (thanks to Thomas Hauser)
# Note: better to use ld=f95 when linking
fortran_compilers['NAGWare Fortran 95'] = {
    'versmatch' : r'NAGWare Fortran 95 compiler Release (?P<version>[^\s]*)',
    'FFLAGS' : '',
    'FOPT' : '-O4',
    'FDEBUG' : '-gline -g90 -g -nan',
    'NOTE' : 'use LD=f95 when linking',
    'COMPILERULES':'%.o : %.f90 ; $(FC) -c $(FFLAGS) $< -o $@'
    }
# Compaq Fortran V1.0-920 on Linux 2.2.13-0.9 alpha (pearu)
# Note: fort needs LD=fort when linking!
fortran_compilers['Compaq Fortran']={
    'versmatch':r'Compaq Fortran (?P<version>[^\s]*)',
    'FFLAGS':'-assume no2underscore -nomixed_str_len_arg',
    'CFLAGS':'-DUSESCOMPAQFORTRAN',
    'FOPT':'-O4 -align dcommons -arch host -assume bigarrays -assume nozsize -math_library fast -tune host',
    'FDEBUG':'-g -check_bounds',
    'NOTE':'You must use LD=fort for linking codes produced with Compaq Fortran compiler.',
    'LIBS':"-lUfor -lfor -lFutil -lcpml -lots -lc",
    'COMPILERULES':'%.o : %.f90 ; $(FC) -c $(FFLAGS) $< -o $@'
    }
fortran_compilers['g77 3.x.x']={
    'versmatch':r'g77 version (?P<version>3[^\s*]*)',
    'FFLAGS':'-fpic -Wall',
    #'LIBS(GNU ld)':"-L$(shell gcc -v 2>&1 | grep specs | sed -e 's/Reading specs from //g' | sed -e 's/\/specs//g') -lg2c",
    'LIBS':"-L$(shell gcc -v 2>&1 | grep specs | sed -e 's/Reading specs from //g' | sed -e 's/\/specs//g') -lg2c -lgcc",
    'NOTE' : 'use LD=gcc when linking',
    }
fortran_compilers['g77 2.x.x']={
    'versmatch':r'g77 version (?P<version>2[^\s*]*)',
    'FFLAGS':'-fpic -Wall',
    #'LIBS(GNU ld)':"-L$(shell gcc -v 2>&1 | grep specs | sed -e 's/Reading specs from //g' | sed -e 's/\/specs//g') -lg2c",
    'LIBS':"-L$(shell gcc -v 2>&1 | grep specs | sed -e 's/Reading specs from //g' | sed -e 's/\/specs//g') -lg2c -lgcc",
    'NOTE' : 'use LD=gcc when linking',
    }
fortran_compilers['gcc 2.x.x']={
    'versmatch':r'gcc version (?P<version>2[^\s*]*)',
    'FFLAGS':'-fpic -Wall',
    #'LIBS(GNU ld)':"-L$(shell gcc -v 2>&1 | grep specs | sed -e 's/Reading specs from //g' | sed -e 's/\/specs//g') -lg2c",
    'LIBS':"-L$(shell gcc -v 2>&1 | grep specs | sed -e 's/Reading specs from //g' | sed -e 's/\/specs//g') -lg2c -lgcc",
    'FOPT':'-O3 -Wall -fast',
    }
fortran_compilers['egcs']={
    'versmatch':r'gcc version (?P<version>egcs[^\s*]*)',
    'FFLAGS':'-fpic -Wall',
    #'LIBS(GNU ld)':"-L$(shell egcs -v 2>&1 | grep specs | sed -e 's/Reading specs from //g' | sed -e 's/\/specs//g') -lg2c",
    'LIBS':"-L$(shell egcs -v 2>&1 | grep specs | sed -e 's/Reading specs from //g' | sed -e 's/\/specs//g') -lg2c",
    'FOPT':'-O3 -Wall',
    }
fortran_compilers['Absoft F77']={
    'versmatch':r'FORTRAN 77 Compiler (?P<version>[^\s*,]*).*?Absoft Corp',
    'LIBS':'-L$$ABSOFT/lib -lfio -lf77math',
    'FFLAGS':'-B108 -N22 -N90 -N110',
#    'LIBS(GNU ld)':"-L$(shell gcc -v 2>&1 | grep specs | sed -e 's/Reading specs from //g' | sed -e 's/\/specs//g') -lg2c",
    'FOPT':'-O -B101',
    }
fortran_compilers['Absoft F90']={
    'versmatch':r'f90:.*?Absoft.*?Version (?P<version>[^\s*,]*)',
    'LIBS':'-L$$ABSOFT/lib -lfio -lf77math -lf90math',
    'FFLAGS':'-B108 -YCFRL=1 -YCOM_NAMES=LCS -YCOM_PFX -YCOM_SFX=_ -YEXT_PFX -YEXT_NAMES=LCS',
#    'LIBS(GNU ld)':"-L$(shell egcs -v 2>&1 | grep specs | sed -e 's/Reading specs from //g' | sed -e 's/\/specs//g') -lg2c",
    'FOPT':'-O -B101',
    'COMPILERULES':'%.o : %.f90 ; $(FC) -c $(FFLAGS) $< -o $@\n%.o : %.f95 ; $(FC) -c $(FFLAGS) $< -o $@',
    }
# VAST/f90 Fortran 90 Compiler: Personal V3.4N5
fortran_compilers['VAST F90']={
    'versmatch':r'\s*Pacific-Sierra Research vf90 Personal\s+(?P<version>V[\d.\w]+)',
    'FOPT':'-O',
    'COMPILERULES':'%.o : %.f90 ; $(FC) -c $(FFLAGS) $<\n%.o : %.f ; $(FC) -c $(FFLAGS) $<',
    'LIBS':"-L$(shell dirname `which f90`) -lvast90 -L$(shell gcc -v 2>&1 | grep specs | sed -e 's/Reading specs from //g' | sed -e 's/\/specs//g') -lg2c",
    'F2PYFLAGS':'--external-modroutines',
    }
# MIPSpro Compilers: Version 7.30
fortran_compilers['MIPSpro 7 Compilers']={
    'versmatch':r'MIPSpro Compilers: Version (?P<version>7[^\s*]*)',
    #'FFLAGS':"-$(shell test -f /etc/compiler.defaults && echo `cat /etc/compiler.defaults | sed -e 's/:/\\n/g'` | grep isa= | sed -e 's/isa=//g' || echo mips4) -r8 -n32",
    'FFLAGS':"-mips4 -n32",
    'FOPT':'-O3',
    'FDEBUG':'-g',
    'COMPILERULES':'%.o : %.f90 ; $(FC) -c $(FFLAGS) $< -o $@'
    }
c_compilers['MIPSpro 7 Compilers']={
    'versmatch':r'MIPSpro Compilers: Version (?P<version>7[^\s*]*)',
    'CFLAGS':"-mips4 -n32 -DF2PY_CB_RETURNCOMPLEX",
    'COPT':'-O3',
    'CDEBUG':'-g'
    }
linkers['Linker for MIPSpro 7 Compilers']={
    'versmatch':r'ld32: INFO[^:]*: Version (?P<version>7[^\s*]*)',
    'LIBS':'-L/usr/lib32 -lftn -lc -lfortran',
    'LDSHARED':'-shared -s'
    }
#Portland Group Fortan 90 compiler on linux (Mikael P Johansson)
fortran_compilers['Portland Group Fortran 90'] = {
        'versmatch' : r'pgf90 (?P<version>[^\s]*)',
        'FFLAGS' : '',
        'FOPT'   : '-fast',
        'FDEBUG' : '-g -Mbounds',
        'LIBS'   : "-L${PGI}/linux86/lib -lpgf90 -lpgf90_rpm1 -lpgf902 -lpgf90rtl -lpgftnrtl -lpgc -lc",
        'COMPILERULES' : '%.o : %.f90 ; $(FC) -c $(FFLAGS) $< -o $@'
        }

# Compaq C V6.2-002 on Linux 2.2.13-0.9 alpha (pearu)
c_compilers['Compaq C']={
    'versmatch':r'Compaq C (?P<version>[^\s]*)', 
    'CFLAGS':'-fpic',
    'COPT':'-O -fast'
    }

c_compilers['gcc 3.x.x']={
    'versmatch':r'gcc version (?P<version>3[^\s*]*)',
    'CFLAGS':'-fpic -Wall',
    }
# g77 version 2.95.2 on Linux 2.2.13-0.9 alpha (pearu)
c_compilers['g77 2.x.x']={
    'versmatch':r'g77 version (?P<version>2[^\s*]*)',
    'CFLAGS':'-fpic -Wall',
    }
c_compilers['gcc 2.x.x']={
    'versmatch':r'gcc version (?P<version>2[^\s*]*)',
    'CFLAGS':'-fpic -Wall',
    }
c_compilers['egcs']={
    'versmatch':r'gcc version (?P<version>egcs[^\s*]*)',
    'CFLAGS':'-fpic -Wall',
    'COPT':'-O3 -Wall'
    }
linkers['NAGWare Fortran'] = { # thanks to Thomas Hauser
     'versmatch' : r'NAGWare Fortran 95 compiler Release (?P<version>[^\s]*)',
     'LDFLAGS' : '-Wl,-shared'
     }
linkers['Compaq Fortran']={
    'versmatch':r'Compaq Fortran (?P<version>[^\s]*)',
    'LDFLAGS':'-Wl,-s -nofor_main',
    'LDSHARED':'-shared',
    }
linkers['g77']={
    'versmatch':r'g77 version (?P<version>[^\s*]*)',
    'LDFLAGS':'-Wl,-s',
    'LDSHARED':'-shared',
    }
linkers['gcc']={
    'versmatch':r'gcc version (?P<version>[^\s*]*)',
    'LDFLAGS':'-Wl,-s',
    'LDSHARED':'-shared',
    'LIBS':'-lg2c -lgcc'
    }
# GNU ld version 2.9.5 (with BFD 2.9.5.0.22) on Linux 2.2.13-0.9 alpha (pearu)
linkers['GNU ld']={
    'versmatch':r'GNU ld version (?P<version>[^\s*]*)',
    'LDSHARED':'-shared -s',
    'LIBS':'@LIBS(GNU ld)@',
    'LIBS(GNU ld)':''
    }
# ld: 92453-07 linker ld B.10.24 on HP-UX B.10.20 A 9000/780 (pearu)
linkers['HP-UX ld']={
    'versmatch':r'ld: (?P<version>.*?B[.]10.*)',
    'LDSHARED':'-b',
    }
# HP fortran 77
fortran_compilers['HP-UX fort77']={
    'versmatch':r'(?P<version>.*?LPATH.*)',
    'FFLAGS':'+DAportable -g +T +ppu +z +e',
    'FOPT':'-O',
    'FDEBUG':'-g',
    'CFLAGS':'-DF2PY_CB_RETURNCOMPLEX',
    'LIBS':' -lU77 -lf'
    }
# HP fortran 77
fortran_compilers['HP-UX f77']={ 
    'versmatch':r'f77 -v: (LPATH is *)', # is this correct?
    'FFLAGS':'+DAportable -g +T +ppu +z +e',
    'FOPT':'-O',
    'FDEBUG':'-g',
    'CFLAGS':'-DF2PY_CB_RETURNCOMPLEX',
    'LIBS':' -lU77 -lf',
    }

makefile_template="""\
# -*- makefile -*-
# File: Makefile-@modulename@
# Usage:
#     make -f Makefile-@modulename@ [MODE=opt|debug]
# Notes:
# 1) You must use GNU make; try `gmake ..' if `make' fails.
# 2) This file is auto-generated with f2py (version @f2py_version@).
#    f2py is a Fortran to Python Interface Generator (FPIG), Second Edition,
#    written by Pearu Peterson <pearu@ioc.ee>.
#    See http://cens.ioc.ee/projects/f2py2e/
#    Generation date: """+time.asctime(time.localtime(time.time()))+"""
#    $R"""+"""evision:$
#    $D"""+"""ate:$

# Recommendation notes produced by f2py2e/buildmakefile.py:
# *** @NOTE@

PYINC = @PYINC@
INCLUDES = @INCLUDES@
LIBS = @LIBS@
PYTHON = """+sys.executable+"""

# Wrapper generator:
F2PY = @F2PY@

# Fortran compiler: @FC@ (@FC_version@)
FC = @FC@
FOPT = @FOPT@
FDEBUG = @FDEBUG@
FFLAGS = @FCSHARED@ @FFLAGS@ 

# C compiler: @CC@ (@CC_version@)
CC = @CC@
COPT = @COPT@
CDEBUG = @CDEBUG@
CFLAGS = @CCSHARED@ @CFLAGS@

# Linker: @LD@ (@LD_version@)
LD = @LD@
LDFLAGS = @LDSHARED@ @LDFLAGS@
SO = @SO@

ifeq '$(MODE)' 'debug'
FFLAGS += $(FDEBUG)
CFLAGS += $(CDEBUG)
endif
ifeq '$(MODE)' 'opt'
FFLAGS += $(FOPT)
CFLAGS += $(COPT)
endif
FFLAGS += $(INCLUDES)
CFLAGS += $(PYINC) $(INCLUDES)

SRCC = fortranobject.c @SRCC@
SRCF = @SRCF@
SRCS = $(SRCC) $(SRCF)
OBJC = $(filter %.o,$(SRCC:.c=.o) $(SRCC:.cc=.o) $(SRCC:.C=.o))
OBJF = $(filter %.o,$(SRCF:.f90=.o) $(SRCF:.F90=.o) $(SRCF:.f=.o) $(SRCF:.F=.o) $(SRCF:.f77=.o) $(SRCF:.F77=.o) $(SRCF:.for=.o) $(SRCF:.FOR=.o))
OBJS = $(OBJC) $(OBJF)

INSTALLNAME = f2py-apps
INSTALLDIRECTORY = @PYLIB@/site-packages/$(INSTALLNAME)
INSTALLDIR = install -d -c
INSTALLEXEC = install -m 755 -c

all: @modulename@

@modulename@: @modulename@module$(SO)
@modulename@module$(SO) : @modulename@module.o $(OBJS)
\t$(LD) $(LDFLAGS) -o $@ $< $(OBJS) $(LIBS)

@modulename@module.o: @modulename@module.c
@modulename@module.c: @signaturefile@
\t$(F2PY) @signaturefile@ $(F2PYFLAGS)

@signaturefile@: $(SRCF)
\ttest -f @signaturefile@ || $(F2PY) -m @modulename@ -h @signaturefile@ $(SRCF) $(F2PYFLAGS)

$(OBJS) : $(SRCS)

@COMPILERULES@

test: @modulename@module$(SO)
\t$(PYTHON) -c 'import @modulename@;print @modulename@.__doc__'
install: @modulename@module$(SO)
\t$(INSTALLDIR) $(INSTALLDIRECTORY)
\t$(INSTALLEXEC) @modulename@module$(SO) $(INSTALLDIRECTORY)
\tcd $(INSTALLDIRECTORY) && echo \"$(INSTALLNAME)\" > ../$(INSTALLNAME).pth

.PHONY: clean distclean debug test install @modulename@
debug:
\techo \"OBJS=$(OBJS)\"
\techo \"SRCS=$(SRCS)\"
clean:
\t$(RM) *.o *.mod core @modulename@module.{dvi,log} $(OBJS)
distclean: clean
\t$(RM) *.so *.sl @modulename@module.{tex,so,c} @modulename@-f2pywrappers.f
\t$(RM) .f2py_get_compiler_* *-raw.pyf fortranobject.{c,h}
"""

def showcompilers(options):
    global PATH
    try: PATH=os.environ['PATH']
    except: PATH='';sys.stderr.write('\tWarning: no PATH in environment?!\n')
    tmp,PATH=PATH,[]
    for s in string.split(tmp,':'):
        if not s in PATH: PATH.append(s)
    get_cc(options)
    get_fc(options)
    get_ld(options)
def build(module,options):
    global modulename,PATH
    modulename=module['name']
    fn=os.path.join(options['buildpath'],'Makefile-%s'%modulename)
    if os.path.isfile(fn):
        if '--overwrite-makefile' in options['makefileopts']:
            outmess('Creating %s...\n'%`fn`)
            outmess('\tFile %s exists: creating backup %s.\n'%(`fn`,`fn+'.bak'`))
            import shutil
            shutil.copyfile(fn,fn+'.bak')
        else:
            outmess('Keeping existing %s. Use --overwrite-makefile to overwrite.\n'%(`fn`))
            return
    else:
        outmess('Creating %s...\n'%`fn`)
    try: PATH=os.environ['PATH']
    except: PATH='';sys.stderr.write('\tWarning: no PATH in environment?!\n')
    tmp,PATH=PATH,[]
    for s in string.split(tmp,':'):
        if not s in PATH: PATH.append(s)
    rd={'modulename':modulename,
        'f2py_version':f2py_version,
        'F2PY':sys.argv[0],
        'signaturefile':get_signaturefile(options),
        'separatorsfor':{'NOTE':'\n# *** '}
        }
    for s in ['LDFLAGS','LDSHARED','CFLAGS','COPT','FFLAGS','FOPT','CDEBUG','FDEBUG','NOTE',
              'COMPILERULES']:
        rd[s]=[]
    for fnc in [
                get_fc,
                get_ld,
                get_cc,
                get_so,
                get_srcf,get_srcc,
                get_pyinc,
                get_libs,
                get_fcshared,
                get_fflags,
                get_ccshared,
                get_cflags,
                get_opts,
                checkPyNumVersion
                ]:
        rd=dictappend(rd,fnc(options))
    needwrap = containscommon(module) or containsmodule(module)
    if not needwrap:
        for n in module['interfaced']: # determine whether Fortran function wrappers are required
            nb=None
            for bi in module['body']:
                for b in bi['body']:
                    if b['name']==n: nb=b;break
            if not nb: continue
            if isfunction_wrap(nb):
                needwrap = 1
                break
    #if not needwrap:
    #    for b in module['body']:
    #        if l_or(isfunction,issubroutine)(b):
    #            needwrap = 1
    #            break
    if needwrap:
        rd['SRCF'].append('%s-f2pywrappers.f'%modulename)
        rd['COMPILERULES'].append('\n\n%s-f2pywrappers.f:\n\techo "C\\n" > %s-f2pywrappers.f'%(modulename,modulename))

    f=open(fn,'w')
    f.write(replace(makefile_template,rd,' '))
    f.close()
    
def checkPyNumVersion(options):
    ret={}
    try:
        import Numeric
    except ImportError:
        outmess('Warning: Failed to import Numeric?! You need Numeric 13 or higher\n')
    try:
        v = Numeric.__version__
        outmess('Numeric version is %s\n'%(v))
        v = string.split(v,'.')[0]
        if eval(v)<13:
            outmess('Warning: Numeric version 13 or higher is required but got %s.\n'%v)
            ret['NOTE']='Warning: You need to install Numeric version 13 or higher.'
            ret['CFLAGS']='-DOLDPYNUM'
    except:
        print sys.exc_type,':',sys.exc_value
        outmess('Warning: Failed to determine PyNum version.\n')
    return ret
def get_signaturefile(options):
    global modulename
    if options.has_key('signsfile'):
        return options['signsfile']
    else:
        return modulename+'-raw.pyf'
def get_pyinc(options):
    pyinc=[os.path.join(sys.prefix,'include','python'+sys.version[:3])]
    pylib=os.path.join(sys.prefix,'lib','python'+sys.version[:3])
    inc=[]
    for o in options['makefileopts']:
        if o[:2]=='-I': inc.append(o)
    fl=1
    tmp=[]
    for d in ['Numeric','numeric','']:
        if os.path.isfile(os.path.join(pyinc[0],d,'arrayobject.h')):
            if d: tmp.append(os.path.join(pyinc[0],d))
            if os.access(os.path.join(pyinc[0],d,'arrayobject.h'),os.R_OK):
                fl=0;break
            else:
                outmess('\tWarning: %s is not readable.\n'%os.path.join(pyinc[0],d,'arrayobject.h'))
    tmp.reverse()
    pyinc=pyinc+tmp
    pyinc.reverse()
    if fl: outmess('\tWarning: Could not find the location of arrayobject.h.\n')
    return {'PYINC':reduce(lambda a,b:'%s -I%s'%(a,b),pyinc,''),
            'INCLUDES':inc,
            'PYLIB':pylib}
def get_srcf(options):
    srcf=[]
    for o in options['makefileopts']:
        if re.match(r'.*?[.](f|F|for|FOR|F90|f90|f95|F95)\Z',o):
            srcf.append(o)
    return {'SRCF':srcf}
def get_srcc(options):
    srcc=[]
    for o in options['makefileopts']:
        if re.match(r'.*?[.](c|cc|C|c[+][+]|cpp)\Z',o):
            srcc.append(o)
    return {'SRCC':srcc}

def get_fc(options):
    return get_compiler(options,'Fortran compiler','FC','f77',
                        ['FC','F77','FORT','F90','F95'],
                        r'FC=(?P<name>.*)',
                        ['pgf90','fort','fort77','f77','g77','f90','f95','f77fe',
                         'ifc'],
                        fortran_compilers)
def get_cc(options):
    return get_compiler(options,'C compiler','CC','cc',
                        ['CC'],
                        r'CC=(?P<name>.*)',
                        ['cc','CC','gcc','ccc','g++','c++','egcs'],
                        c_compilers)
def get_ld(options):
    #LDSHARED=string.split(LDSHARED)
    #LD=LDSHARED[0]
    #LDSHARED=string.join(LDSHARED[1:],' ')
    return get_compiler(options,'Linker','LD','ld',
                        ['LD'],
                        r'LD=(?P<name>.*)',
                        ['ld','gcc','egcs','fort','f95','ifc'],
                        linkers)

def get_so(options):
    return get_variable(options,'SO','SO','.so',
                        ['SO'],
                        r'S0=(?P<name>.*)',
                        'distutils.sysconfig')
def get_fcshared(options):
    return get_variable(options,'FCSHARED','FCSHARED','',
                        ['FCSHARED'],
                        r'FCSHARED=(?P<name>.*)',
                        '')
def get_fflags(options):
    return get_variable(options,'FFLAGS','FFLAGS','',
                        ['FFLAGS'],
                        r'FFLAGS=(?P<name>.*)',
                        '')
def get_ccshared(options):
    return get_variable(options,'CCSHARED','CCSHARED','',
                        ['CCSHARED'],
                        r'CCSHARED=(?P<name>.*)',
                        '')
def get_cflags(options):
    ret=get_variable(options,'CFLAGS','CFLAGS','',
                     ['CFLAGS'],
                     r'CFLAGS=(?P<name>.*)',
                     '')
    for o in options['makefileopts']:
        if o[:2] in ['-D','-U']:
            ret=dictappend(ret,{'CFLAGS':o})
    return ret
def get_libs(options):
    ret=get_variable(options,'LIBS','LIBS','',
                     ['LIBS'],
                     r'LIBS=(?P<name>.*)',
                     '')
    for o in options['makefileopts']:
        if o[:2] in ['-l','-L']:
            ret=dictappend(ret,{'LIBS':o})
    return ret
def get_opts(options):
    ret={}
    for s in ['COPT']:
        ret=dictappend(ret,get_variable(options,s,s,'',
                                    [s],
                                    '%s=(?P<name>.*)'%s,
                                    ''))
    return ret
def get_f2pyflags(options):
    ret={}
    for env in ['FC']:
        sfn = os.path.join(options['buildpath'],'.f2py_get_compiler_%s'%(env))
        try:
            f = open(sfn,'r')
            ret2 = eval(f.read())
            f.close()
            if ret2.has_key('F2PYFLAGS'):
                ret=dictappend(ret,{'F2PYFLAGS':ret2['F2PYFLAGS']})
        except: pass
    ret2 = dictappend(ret,get_variable(options,'F2PYFLAGS','F2PYFLAGS','',
                        ['F2PYFLAGS'],
                        r'F2PYFLAGS=(?P<name>.*)',
                        ''))['F2PYFLAGS']
    ret = {}
    for r in ret2:
        if r=='--external-modroutines': ret[r]=1
        elif r=='--no-external-modroutines': ret['--external-modroutines']=0
    return ret

# Generic functions
def get_compiler(options,name,env,defenv,envs,enveq,envtry,compilers):
    sfn = os.path.join(options['buildpath'],'.f2py_get_compiler_%s'%(env))
    try:
        f = open(sfn,'r')
        ret = eval(f.read())
        f.close()
        outmess('\t%s: %s (from %s)\n'%(name,replace('@%s@ (@%s_version@)'%(env,env),ret),sfn))
        return ret
    except:
        if os.path.isfile(sfn):
            outmess('\tWarning: File %s exists but failed to use its contents.\n'%(sfn))
    try:
        import distutils.sysconfig
        exec '%s=distutils.sysconfig.%s'%(env,env)
        fl=1
    except:
        fl=0
        exec '%s=%s'%(env,`defenv`)
    if not get_existingprogs([eval(env)]):
        outmess('\tWarning: %s %s is not in PATH. Will try other names.\n'%(name,`eval(env)`))
        fl=1
    for s in envs:
        if os.environ.has_key(s):
            exec '%s,fl=%s,0'%(env,`os.environ[s]`)
            break
    if fl==0 and not get_existingprogs([eval(env)]):
        outmess('\tWarning: %s %s is not in PATH. Will try other names.\n'%(name,`eval(env)`))
        fl=1
    for o in options['makefileopts']:
        m=re.match(enveq,o)
        if m:
            exec('%s,fl=%s,0'%(env,`m.group('name')`))
            break
    if fl==0 and not get_existingprogs([eval(env)]):
        outmess('\tWarning: %s %s is not in PATH. Will try other names.\n'%(name,`eval(env)`))
        fl=1
    if fl:
        comps=get_existingprogs(envtry)
        if comps:
            exec('%s=%s'%(env,`comps[0]`))
        else:
            outmess('\tWarning: could not find %s. Using %s.\n'%(name,`eval(env)`))
    ret={}
    ret[env]=eval(env)
    ret[env+'_version']='???'
    if not get_existingprogs([eval(env)]):
        outmess('\tWarning: %s %s is not in PATH.\n'%(name,`eval(env)`))
    else:
        comp,vers='',''
        errcomp,errvers='',''
        for v in ['-v','-V','-version',
                  '-c `echo -e "\tsubroutine dummy()\n\tend" > /tmp/f2py_dummy.f;echo /tmp/f2py_dummy.f` -o /tmp/f2py_dummy_dev_null',
                  '-c /tmp/f2py_dummy.f',
                  '-V -c /tmp/f2py_dummy.f -o /tmp/f2py_dummy_dev_null',
                  #'-c `echo -e "\tsubroutine dummy()\n\tend" > /tmp/f2py_dummy.f;echo /tmp/f2py_dummy.f`'
                  ]:
            o=commands.getstatusoutput('%s %s 2>&1'%(eval(env),v))
            if not o[0]:
                for l in string.split(o[1],'\n'):
                    for c in compilers.keys():
                        m = re.match(compilers[c]['versmatch'],l)
                        if m:
                            comp,vers=c,m.group('version')
                            break
                    if comp:break
            elif not errcomp:
                for l in string.split(o[1],'\n'):
                    for c in compilers.keys():
                        m = re.match(compilers[c]['versmatch'],l)
                        if m:
                            errcomp,errvers=c,m.group('version')
                            break
        if not comp:
            comp,vers=errcomp,errvers
        if comp:
            ret[env+'_version']=`comp`+' '+vers
            ret=dictappend(ret,compilers[comp])
    outmess('\t%s: %s\n'%(name,replace('@%s@ (@%s_version@)'%(env,env),ret)))
    f = open(sfn,'w')
    f.write('# This file is created by f2py (version %s). You should not edit this file manually.\n'%(f2py_version))
    f.write('# But you may want to remove this file if using different compiler/linker.\n')
    pprint.pprint(ret,f)
    f.close()
    return ret

def get_variable(options,name,env,defenv,envs,enveq,tryfrom):
    try:
        exec 'import %s'%tryfrom
        exec '%s=%s.%s'%(env,tryfrom,env)
        fl=1
    except:
        fl=0
        exec '%s=%s'%(env,`defenv`)
    for s in envs:
        if os.environ.has_key(s):
            exec '%s,fl=%s,0'%(env,`os.environ[s]`)
            break
    if options.has_key('makefileopts'):
        for o in options['makefileopts']:
            m=re.match(enveq,o)
            if m:
                exec('%s,fl=%s,0'%(env,`m.group('name')`))
                break
    ret={}
    ret[env]=eval(env)
    #outmess('\tenv: %s=%s\n'%(name,eval(env)))
    return ret
def get_existingprogs(progsin):
    global PATH
    progs=[]
    progsin=map(lambda p:string.split(p)[0],progsin)
    for p in PATH:
        for prog in progsin:
            fp=os.path.join(p,prog)
            if os.path.isfile(fp) and not prog in progs:
                progs.append(prog)
    return progs
def replace(str,dict,defaultsep=''):
    if type(dict)==types.ListType:
        return map(lambda d,f=replace,sep=defaultsep,s=str:f(s,d,sep),dict)
    for k in 2*dict.keys():
        if k=='separatorsfor': continue
        if dict.has_key('separatorsfor') and dict['separatorsfor'].has_key(k):
            sep=dict['separatorsfor'][k]
        else:
            sep=defaultsep
        if type(dict[k])==types.ListType:
            str=string.replace(str,'@%s@'%(k),string.join(flatlist(dict[k]),sep))
        else:
            str=string.replace(str,'@%s@'%(k),dict[k])
    return str

