#!/usr/bin/env python
"""

buildsetup --- builds setup_<modulename>.py (see function build() below).

    Below are defined some Fortran and C compilers, and linkers (obsolete).
    New ones should be added to fortran_support.

Copyright 2001 Pearu Peterson all rights reserved,
Pearu Peterson <pearu@cens.ioc.ee>          
Permission to use, modify, and distribute this software is given under the
terms of the LGPL.  See http://www.fsf.org

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
$Date: 2002/01/11 16:34:16 $
Pearu Peterson
"""

__version__ = "$Revision: 1.9 $"[10:-1]

import types,string,sys,os,time,re,commands,pprint

from auxfuncs import flatlist,dictappend,isfunction_wrap,isfunction,\
     issubroutine,l_or,containscommon,containsmodule
outmess=sys.stdout.write
import __version__
f2py_version = __version__.version

show = pprint.pprint

manifest_in_template="""\
include @signaturefile@
include @modulename@.pyf
include @modulename@.tex
include fortranobject.h
include fortranobject.c
include @modulename@module.c
@includeSRCC@
@includeSRCF@
"""

setup_template="""\
#!/usr/bin/env python
# File: setup_@modulename@.py
# Usage:
#    python setup_@modulename@.py build
#    python setup_@modulename@.py install
# Notes:
# 1) This file is auto-generated with f2py (version @f2py_version@).
#    f2py is a Fortran to Python Interface Generator (FPIG), Second Edition,
#    written by Pearu Peterson <pearu@cens.ioc.ee>.
#    See http://cens.ioc.ee/projects/f2py2e/
#    Generation date: """+time.asctime(time.localtime(time.time()))+"""
# 2) This setup file requires fortran_support module being installed.
#    fortran_support is included with f2py2e (originally copied from SciPy).
#
#    $R"""+"""evision:$
#    $D"""+"""ate:$

# Recommendation notes produced by f2py2e/buildsetup.py:
# *** @NOTE@

import string

modulename = '@modulename@'

libraries = [@libraries@]
library_dirs = [@library_dirs@]
extra_objects = []
extra_link_args = string.split('@LIBS@')
extra_compile_args = []
include_dirs = [@INCLUDES@]
define_macros = @define_macros@
undef_macros = @undef_macros@
c_files = ['@modulename@module.c','fortranobject.c']+[@SRCC@]
fortran_files = [@SRCF@]
#c_files = map(lambda s:s,c_files)
#fortran_files = map(lambda s:s,fortran_files)
flibraries = [(modulename,{'sources': fortran_files})]

from distutils.core import setup, Extension

ext = Extension(modulename,
                sources = c_files,
                include_dirs = include_dirs,
                libraries = libraries,
                library_dirs = library_dirs,
                extra_objects = extra_objects,
                extra_compile_args = extra_compile_args,
                extra_link_args = extra_link_args,
                define_macros = define_macros,
                undef_macros = undef_macros
                )

ext_modules = [ext]

if __name__ == '__main__':
    import fortran_support

    setup(name = modulename,
      # You may want to change the following items
      version = '0.1',
      author = 'f2py generated',
      #author_email = 'you@some.where.com',
      #maintainer = 'your name',
      #maintainer_email = 'your email',
      licence = 'LGPL',
      url = 'http://cens.ioc.ee/projects/f2py2e/',
      description = '@modulename@ - description',

      distclass = fortran_support.FortranDistribution,
      cmdclass = {'build': fortran_support.build_fortran,
                  'build_flib': fortran_support.build_flib,
                  'build_ext': fortran_support.build_ext_fortran,
                  'build_clib':fortran_support.build_clib_new.build_clib,
                  'sdist':fortran_support.sdist_fortran,
                  },    
      fortran_libraries = flibraries,
      ext_modules =  ext_modules,
      packages = [modulename],
      package_dir = {modulename:'.'}
      )
"""

scipy_setup_template="""\
#!/usr/bin/env python
# File: setup_@modulename@.py
# Usage:
#    python setup_@modulename@.py build
#    python setup_@modulename@.py install
# Notes:
# 1) This file is auto-generated with f2py (version @f2py_version@).
#    f2py is a Fortran to Python Interface Generator (FPIG), Second Edition,
#    written by Pearu Peterson <pearu@cens.ioc.ee>.
#    See http://cens.ioc.ee/projects/f2py2e/
#    Generation date: """+time.asctime(time.localtime(time.time()))+"""
# 2) This setup file requires scipy_distutils module being installed.
#    scipy_distutils is included with f2py2e (originally copied from SciPy).
#
#    $R"""+"""evision:$
#    $D"""+"""ate:$

import string

libraries = [@libraries@]
library_dirs = [@library_dirs@]
extra_objects = []
extra_link_args = string.split('@LIBS@')
extra_compile_args = []
include_dirs = [@INCLUDES@]
define_macros = @define_macros@
undef_macros = @undef_macros@
c_files = ['@modulename@module.c']+[@SRCC@]
fortran_files = [@SRCF@]

def setup_package(modulename = '@modulename@'):
    global libraries
    global library_dirs
    global extra_objects
    global extra_link_args
    global extra_compile_args
    global include_dirs
    global define_macros
    global undef_macros
    global c_files
    global fortran_files

    import os
    from scipy_distutils.core import setup, Extension
    from scipy_distutils.misc_util import get_path

    path = get_path(__name__)
    old_path = os.getcwd()
    os.chdir(path)

    try:        
        ext = Extension(modulename,
                        sources = c_files,
                        include_dirs = include_dirs,
                        libraries = libraries,
                        library_dirs = library_dirs,
                        extra_objects = extra_objects,
                        extra_compile_args = extra_compile_args,
                        extra_link_args = extra_link_args,
                        define_macros = define_macros,
                        undef_macros = undef_macros
                        )
        setup(name = modulename,
              # You may want to change the following items
              version = '0.1',
              author = 'f2py generated',
              #author_email = 'you@com',
              #licence = 'LGPL',
              #url = 'http://',
              description = '@modulename@ - description',
              fortran_libraries = [(modulename,{'sources': fortran_files})],
              ext_modules =  [ext],
              packages =     [modulename],
              package_dir = { modulename: '.'}
              )
    finally:
        os.chdir(old_path)

if __name__ == '__main__':
    setup_package()

if __name__ == '__builtin__': # for run_setup in distutils.core
    setup_package()

"""

def build(module,options):
    global modulename
    modulename=module['name']
    fn=os.path.join(options['buildpath'],'setup_%s.py'%modulename)
    mifn=os.path.join(options['buildpath'],'MANIFEST.in')

    if os.path.isfile(fn):
        if '--overwrite-setup' in options['setupopts']:
            outmess('Creating %s...\n'%`fn`)
            outmess('\tFile %s exists: creating backup %s.\n'%(`fn`,`fn+'.bak'`))
            import shutil
            shutil.copyfile(fn,fn+'.bak')
        else:
            outmess('Keeping existing %s. Use --overwrite-setup to overwrite.\n'%(`fn`))
            return
    else:
        outmess('Creating %s...\n'%`fn`)
    rd={'modulename':modulename,
        'f2py_version':f2py_version,
        'F2PY':sys.argv[0],
        'signaturefile':get_signaturefile(options),
        'separatorsfor':{'NOTE':'\n# *** '}
        }
    for s in ['LDFLAGS','LDSHARED','CFLAGS','COPT','FFLAGS','FOPT','CDEBUG','FDEBUG','NOTE',
              'COMPILERULES','define_macros','undef_macros',
              'libraries','library_dirs']:
        rd[s]=[]
    for fnc in [get_srcf,
                get_srcc,
                get_pyinc,
                get_libs,
                get_fcshared,
                get_fflags,
                get_ccshared,
                get_cflags,
                get_opts,
                checkPyNumVersion
                ]:
        rd=dictappend(rd,fnc(options))
    needwrap = containscommon(module) or containsmodule(module)
    if not needwrap:
        for n in module['interfaced']: # determine whether Fortran function wrappers are required
            nb=None
            for bi in module['body']:
                for b in bi['body']:
                    if b['name']==n: nb=b;break
            if not nb: continue
            if isfunction_wrap(nb):
                needwrap = 1
                break
    if needwrap:
        rd['SRCF'] = rd['SRCF'] + "]+['%s-f2pywrappers.f'"%modulename
        rd['includeSRCF'] = rd['includeSRCF'] + "\ninclude %s-f2pywrappers.f"%modulename

    rd['define_macros'] = repr (rd['define_macros'])
    rd['undef_macros'] = repr (rd['undef_macros'])
    f=open(fn,'w')
    if options['doscipysetup']:
        f.write(replace(scipy_setup_template,rd,' '))
    else:
        f.write(replace(setup_template,rd,' '))
    f.close()
    os.chmod(fn,0775)
    if not os.path.isfile(mifn) and options['domanifest']:
        outmess('Creating %s...\n'%`mifn`)
        f=open(mifn,'w')
        f.write(replace(manifest_in_template,rd,' '))
        f.close()

def checkPyNumVersion(options):
    ret={}
    try:
        import Numeric
    except ImportError:
        outmess('Warning: Failed to import Numeric?! You need Numeric 13 or higher\n')
    try:
        v = Numeric.__version__
        outmess('Numeric version is %s\n'%(v))
        v = string.split(v,'.')[0]
        if eval(v)<13:
            outmess('Warning: Numeric version 13 or higher is required but got %s.\n'%v)
            ret['NOTE']='Warning: You need to install Numeric version 13 or higher.'
            ret['CFLAGS']='-DOLDPYNUM'
    except:
        print sys.exc_type,':',sys.exc_value
        outmess('Warning: Failed to determine PyNum version.\n')
    return ret

def get_signaturefile(options):
    global modulename
    if options.has_key('signsfile'):
        return options['signsfile']
    else:
        return os.path.join(options['buildpath'],modulename+'-raw.pyf')

def get_pyinc(options):
    pyinc=[sys.prefix+'/include/python'+sys.version[:3]]
    pylib=sys.prefix+'/lib/python'+sys.version[:3]
    inc=[]
    for o in options['setupopts']:
        if o[:2]=='-I':
            o = o[2:]
            inc.append(o)
    fl=1
    tmp=[]
    for d in ['Numeric','numeric','']:
        if os.path.isfile(os.path.join(pyinc[0],d,'arrayobject.h')):
            if d: tmp.append(os.path.join(pyinc[0],d))
            if os.access(os.path.join(pyinc[0],d,'arrayobject.h'),os.R_OK):
                fl=0;break
            else:
                outmess('\tWarning: %s is not readable.\n'%os.path.join(pyinc[0],d,'arrayobject.h'))
    tmp.reverse()
    pyinc=pyinc+tmp
    pyinc.reverse()
    if fl: outmess('\tWarning: Could not find the location of arrayobject.h.\n')
    for i in range(len(inc)):
        if not os.path.isabs(inc[i]):
            inc[i] = get_relative_path(options['buildpath'],inc[i])
    return {'PYINC':string.join(map(repr,pyinc),','),
            'INCLUDES':string.join(map(repr,inc),','),
            'PYLIB':pylib}

def get_relative_path(base_dir,filename):
    a = string.split(os.path.normpath(os.path.abspath(base_dir)),os.sep)
    b = string.split(os.path.normpath(os.path.abspath(filename)),os.sep)
    for i in range(min(len(a),len(b))):
        if a[i]!=b[i]:
            a,b = a[i:],b[i:]
            break
    ret = os.path.join(*(len(a)*['..']+b))
    return ret

def get_srcf(options):
    srcf=[]
    for o in options['setupopts']:
        if re.match(r'.*?[.](for|f90|f95|ftn|f77|f)\Z',o,re.I):
            o = get_relative_path(options['buildpath'],o)
            srcf.append(o)
    return {'SRCF':string.join(map(repr,srcf),','),
            'includeSRCF':string.join(map(lambda s:'include '+s,srcf),'\n'),
            }

def get_srcc(options):
    srcc=[]
    for o in options['setupopts']:
        if re.match(r'.*?[.](c|cc|c[+][+]|cpp)\Z',o,re.I):
            o = get_relative_path(options['buildpath'],o)
            srcc.append(o)
    return {'SRCC':string.join(map(repr,srcc),','),
            'includeSRCC':string.join(map(lambda s:'include '+s,srcc),'\n'),
            }

def get_fcshared(options):
    return get_variable(options,'FCSHARED','FCSHARED','',
                        ['FCSHARED'],
                        r'FCSHARED=(?P<name>.*)',
                        '')
def get_fflags(options):
    return get_variable(options,'FFLAGS','FFLAGS','',
                        ['FFLAGS'],
                        r'FFLAGS=(?P<name>.*)',
                        '')
def get_ccshared(options):
    return get_variable(options,'CCSHARED','CCSHARED','',
                        ['CCSHARED'],
                        r'CCSHARED=(?P<name>.*)',
                        '')

def get_cflags(options):
    ret=get_variable(options,'CFLAGS','CFLAGS','',
                     ['CFLAGS'],
                     r'CFLAGS=(?P<name>.*)',
                     '')
    for o in options['setupopts']:
        if o[:2]=='-D':
            ret=dictappend(ret,{'CFLAGS':o})
            if '=' in o:
                n,v=string.split(o[2:],'=',1)
            else:
                n,v=o[2:],None
            ret['define_macros'].append ((n,v))
        elif o[:2]=='-U':
            ret=dictappend(ret,{'CFLAGS':o})
            ret['undef_macros'].append (o[2:])
    return ret

def get_libs(options):
    ret=get_variable(options,'LIBS','LIBS','',
                     ['LIBS'],
                     r'LIBS=(?P<name>.*)',
                     '')
    for o in options['setupopts']:
        if o[:2]=='-l':
            if not ret.has_key('libraries'):
                ret['libraries'] = []
            ret['libraries'].append(o[2:])
        if o[:2]=='-L':
            if not ret.has_key('library_dirs'):
                ret['library_dirs'] = []
            ret['library_dirs'].append(o[2:])
    if ret.has_key('library_dirs'):
        for i in range(len(ret['library_dirs'])):
            if not os.path.isabs(ret['library_dirs'][i]):
                ret['library_dirs'][i] = get_relative_path(options['buildpath'],
                                                           ret['library_dirs'][i])    
    return ret

def get_opts(options):
    ret={}
    for s in ['COPT']:
        ret=dictappend(ret,get_variable(options,s,s,'',
                                    [s],
                                    '%s=(?P<name>.*)'%s,
                                    ''))
    return ret

# Generic functions

def get_variable(options,name,env,defenv,envs,enveq,tryfrom):
    try:
        exec 'import %s'%tryfrom
        exec '%s=%s.%s'%(env,tryfrom,env)
        fl=1
    except:
        fl=0
        exec '%s=%s'%(env,`defenv`)
    for s in envs:
        if os.environ.has_key(s):
            exec '%s,fl=%s,0'%(env,`os.environ[s]`)
            break
    if options.has_key('setupopts'):
        for o in options['setupopts']:
            m=re.match(enveq,o)
            if m:
                exec('%s,fl=%s,0'%(env,`m.group('name')`))
                break
    ret={}
    ret[env]=eval(env)
    #outmess('\tenv: %s=%s\n'%(name,eval(env)))
    return ret

def replace(str,dict,defaultsep=''):
    if type(dict)==types.ListType:
        return map(lambda d,f=replace,sep=defaultsep,s=str:f(s,d,sep),dict)
    for k in 2*dict.keys():
        if k=='separatorsfor': continue
        if dict.has_key('separatorsfor') and dict['separatorsfor'].has_key(k):
            sep=dict['separatorsfor'][k]
        else:
            sep=defaultsep
        if type(dict[k])==types.ListType:
            str=string.replace(str,'@%s@'%(k),string.join(flatlist(dict[k]),sep))
        else:
            str=string.replace(str,'@%s@'%(k),dict[k])
    return str

