#!/usr/bin/env python
"""

f2py2e - Fortran to Python C/API generator. 2nd Edition.
         See __usage__ below.

Copyright 1999--2001 Pearu Peterson all rights reserved,
Pearu Peterson <pearu@censioc.ee>          
Permission to use, modify, and distribute this software is given under the
terms of the LGPL.  See http://www.fsf.org

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
$Date: 2002/03/13 12:37:25 $
Pearu Peterson
"""
__version__ = "$Revision: 1.57 $"[10:-1]

import __version__
f2py_version = __version__.version

import sys,os,string,pprint,shutil,types,re
errmess=sys.stderr.write
outmess=sys.stdout.write
show=pprint.pprint

import crackfortran
import rules
import cb_rules
import common_rules
import auxfuncs
import cfuncs
import capi_maps
import buildmakefile
import buildsetup
import func2subr
import f90mod_rules


__usage__ = """\
Usage:

1) To construct extension module sources:

      f2py [<options>] <fortran files> [[[only:]||[skip:]] \\
                                        <fortran functions> ] \\
                                       [: <fortran files> ...]

2) To compile fortran files and build extension modules:

      f2py -c [<options>, <build_flib options>] <fortran files>

3) To generate signature files:

      f2py -h <filename.pyf> ...< same options as in 1) >

Description: This program generates a Python C/API file (<modulename>module.c)
             that contains wrappers for given fortran functions so that they
             can be called from Python. With -c options, it also builds
             the extension modules.
Options:
  -h <filename>    Write signatures of the fortran routines to file <filename>
                   and exit. You can then edit <filename> and use it instead
                   of <fortran files>.
  -m <modulename>  Name of the module; f2py2e will generate a file
                   <modulename>module.c

  <fortran functions>  Names of fortran routines for which Python C/API
                   functions will be generated. Default is all that are found
                   in <fortran files>.
  <fortran files>  Paths to fortran/signature files that will be scanned for
                   <fortran functions> in order to determine their signatures.
  skip:            Ignore fortran functions that follow.
  only:            Use only fortran functions that follow.
  :                Get back to <fortran files> mode.
  --[no-]lower     Do [not] lower the cases in <fortran files>. By default,
                   --lower is assumed with -h key, and --no-lower without -h key.

  --build-dir <dirname>  all f2py generated files are created in <dirname>.
                   Default build directory is the current directory.
  --overwrite-signature  Overwrite existing signature file.
  --overwrite-makefile  Overwrite existing `Makefile-<modulename>' file.
                   [OBSOLETE]
  --overwrite-setup  Overwrite existing `setup_<modulename>.py' file.
                   [NOT FINISHED]
  --[no-]setup     Create (or not) setup_<modulename>.py. Default is
                   --no-setup. [OBSOLETE]
  --scipy-setup    Create setup_<modulename>.py that uses scipy_distutils.
                   [NOT FINISHED].
  --[no-]manifest  Create (or not) MANIFEST.in. Default is --no-manifest.
  --[no-]makefile  Create (or not) Makefile-<modulename>.
                   Default is --no-makefile. [OBSOLETE]
  --[no-]latex-doc Create (or not) <modulename>module.tex.
                   Default is --no-latex-doc.
  --short-latex    Create 'incomplete' LaTeX document (without commands
                   \\documentclass, \\tableofcontents, and \\begin{document},
                   \\end{document}).

  --debug-capi     Create C/API code that will be very verbose. Use when
                   debugging.
  -include'<includefile>' Add CPP #include statement to the C/API code.
                   <includefile> should be in the format of either
                   `\"filename.ext\"' or `<filename.ext>'.
                   As a result <includefile> will be included just before
                   wrapper functions part in the C/API code.

  -makefile <options>  Give additional information for building files
                   `Makefile-<modulename>' and `setup_<modulename>.py'.
                   The following <options> are recognized:
                   `-L<libpath>', `-l<libname>', `-I<incpath>',
                   `[-D|-U]<cpp-macro>', `<additional files *.{f,f90,..} *.c
                   that need to be compiled>', and makefile variables such
                   as F2PYFLAGS, CC,CFLAGS,COPT,FC,FFLAGS,LDSHARED,LDFLAGS,etc
                   (they'll overwrite default ones found from the environment;
                   see buildmakefile.py). The scope of `-makefile' options ends
                   with `:'. By default all <fortran files> will be included to
                   the list of files that need to be compiled. To disable this
                   feature use `--use-libs' option. [OBSOLUTE]

  --[no-]wrap-functions    Create Fortran subroutine wrappers for Fortran
                   functions. --wrap-functions is default because it ensures
                   maximum portability.
  --[no-]external-modroutines

  -v               Print f2py version ID and exit.
  -pyinc           Print Python include path and exit.

Version:     %s
Requires:    Python 1.5.2 or higher (2.x is supported).
             Numerical Python 13 or higher (20.x is supported)
             scipy_distutils (included to f2py distribution)
License:     LGPL (see http://www.fsf.org)
Copyright 1999 - 2002 Pearu Peterson all rights reserved.
WWW page: http://cens.ioc.ee/projects/f2py2e/"""%(f2py_version)


def scaninputline(inputline):
    files,funcs,skipfuncs,onlyfuncs,debug=[],[],[],[],[]
    f,f2,f3,f4,f5,f6=1,0,0,0,0,0
    dolc=-1
    dosetup = 0
    doscipysetup = 0
    domanifest = 0
    domakefile = 0
    dolatexdoc = 0
    wrapfuncs = 1
    buildpath = '.'
##     do_analyze = 1
    signsfile,modulename=None,None
##     format='f77'
    makefileopts=[]
    options=buildmakefile.get_f2pyflags({'buildpath':buildpath})
    for l in inputline:
        if l=='': pass
        elif l=='only:': f=0
        elif l=='skip:': f=-1
        elif l==':': f=1;f4=0
##         elif l=='-f77': format='f77'
##         elif l=='-f90': format='f90'
##         elif l=='-fix': format='fix'
        elif l[:8]=='--debug-': debug.append(l[8:])
        elif l=='--lower': dolc=1
##         elif l=='--skip-analyze': do_analyze=0
        elif l=='--build-dir': f6=1
        elif l=='--no-lower': dolc=0
        elif l=='--setup': dosetup=1
        elif l=='--scipy-setup':
            doscipysetup=1
            dosetup=1
        elif l=='--no-setup': dosetup=0
        elif l=='--manifest': domanifest=1
        elif l=='--no-manifest': domanifest=0
        elif l=='--makefile': domakefile=1
        elif l=='--no-makefile': domakefile=0
        elif l=='--latex-doc': dolatexdoc=1
        elif l=='--no-latex-doc': dolatexdoc=0
        elif l=='--wrap-functions': wrapfuncs=1
        elif l=='--no-wrap-functions': wrapfuncs=0
        elif l=='--short-latex': options['shortlatex']=1
        elif l=='--overwrite-signature': options['h-overwrite']=1
        elif l=='--overwrite-makefile':
            domakefile=1
            makefileopts.append('--overwrite-makefile')
        elif l=='--overwrite-setup':
            dosetup=1
            makefileopts.append('--overwrite-setup')
        elif l=='--external-modroutines': options[l]=1
        elif l=='--no-external-modroutines': options['--external-modroutines']=0
        elif l=='-h': f2=1
        elif l=='-m': f3=1
        elif l=='--use-libs':
            makefileopts.append(l)
        elif l=='-makefile':
            f4=1
        elif l[:2]=='-v':
            print f2py_version
            sys.exit()
        elif l[:6]=='-pyinc':
            print os.path.join(sys.prefix,'include','python'+sys.version[:3])
            sys.exit()
        elif l=='--show-compilers':
            f5=1
        elif l[:8]=='-include':
            cfuncs.outneeds['userincludes'].append(l[9:-1])
            cfuncs.userincludes[l[9:-1]]='#include '+l[8:]
        elif l[0]=='-':
            if f4:
                makefileopts.append(l)
            else:
                errmess('Unknown option %s\n'%`l`)
        elif f2: f2=0;signsfile=l
        elif f3: f3=0;modulename=l
        elif f6: f6=0;buildpath=l
        elif f4:
            makefileopts.append(l)
        elif f==1:
            try:
                open(l).close()
                files.append(l)
            except IOError,detail:
                errmess('IOError: %s. Skipping file "%s".\n'%(str(detail),l))
        elif f==-1: skipfuncs.append(l)
        elif f==0: onlyfuncs.append(l)
    if not f5 and not files and not modulename:
        print __usage__
        #errmess('Expected files and modulename but failed to got one of them\n')
        sys.exit()
    if not os.path.isdir(buildpath):
        outmess('Creating build directory %s'%(buildpath))
        os.mkdir(buildpath)
    if signsfile:
        signsfile = os.path.join(buildpath,signsfile)
    if signsfile and os.path.isfile(signsfile) and not options.has_key('h-overwrite'):
        errmess('Signature file "%s" exists!!! Use --overwrite-signature to overwrite.\n'%(signsfile))
        sys.exit()
##     if format in ['f77','f90','fix']: options['format']=format
##     else:
##         errmess('Illegal format "%s". Exiting.\n'%(format))
##         sys.exit()
    options['debug']=debug
    if dolc==-1 and not signsfile: options['do-lower']=0
    else: options['do-lower']=dolc
    if modulename: options['module']=modulename
    if signsfile: options['signsfile']=signsfile
    if onlyfuncs: options['onlyfuncs']=onlyfuncs
    if skipfuncs: options['skipfuncs']=skipfuncs
##     options['do_analyze'] = do_analyze
    options['dosetup'] = dosetup
    options['doscipysetup'] = doscipysetup
    options['domanifest'] = domanifest
    options['domakefile'] = domakefile
    options['dolatexdoc'] = dolatexdoc
    options['wrapfuncs'] = wrapfuncs
    if not '--use-libs' in makefileopts:
        makefileopts=makefileopts+files
    options['makefileopts']=makefileopts
    options['setupopts']=makefileopts
    options['buildpath']=buildpath
    if f5:
        buildmakefile.showcompilers(options)
        sys.exit()
    return files,options

def callcrackfortran(files,options):
    rules.options=options
    funcs=[]
##     if options['format']=='f77':
##         crackfortran.strictf77=1
##         crackfortran.sourcecodeform='fix'
##     elif options['format']=='f90':
##         crackfortran.strictf77=0
##         crackfortran.sourcecodeform='free'
##     elif options['format']=='fix':
##         crackfortran.strictf77=0
##         crackfortran.sourcecodeform='fix'
##     else:
##         errmess('callcrackfortran: Unknown format. Exiting.\n');sys.exit()
    crackfortran.debug=options['debug']
    if options.has_key('module'):
        crackfortran.f77modulename=options['module']
    if options.has_key('skipfuncs'):
        crackfortran.skipfuncs=options['skipfuncs']
    if options.has_key('onlyfuncs'):
        crackfortran.onlyfuncs=options['onlyfuncs']
##     if options.has_key('do_analyze'):
##         crackfortran.do_analyze=options['do_analyze']
    crackfortran.dolowercase=options['do-lower']
    postlist=crackfortran.crackfortran(files)
    if options.has_key('signsfile'):
        outmess('Saving signatures to file "%s"\n'%(options['signsfile']))
        pyf=crackfortran.crack2fortran(postlist)
        if options['signsfile']=='stdout':
            sys.stdout.write(pyf)
        else:
            f=open(options['signsfile'],'w')
            f.write(pyf)
            f.close()
    return postlist

def buildmodules(list):
    cfuncs.buildcfuncs()
    outmess('Building modules...\n')
    modules,mnames,isusedby=[],[],{}
    for i in range(len(list)):
        if string.find(list[i]['name'],'__user__')>=0:
            cb_rules.buildcallbacks(list[i])
        else:
            if list[i].has_key('use'):
                for u in list[i]['use'].keys():
                    if not isusedby.has_key(u): isusedby[u]=[]
                    isusedby[u].append(list[i]['name'])
            modules.append(list[i])
            mnames.append(list[i]['name'])
    ret = {}
    for i in range(len(mnames)):
        if isusedby.has_key(mnames[i]):
            outmess('\tSkipping module "%s" which is used by %s.\n'%(mnames[i],string.join(map(lambda s:'"%s"'%s,isusedby[mnames[i]]),',')))
        else:
            um=[]
            if modules[i].has_key('use'):
                for u in modules[i]['use'].keys():
                    if isusedby.has_key(u) and u in mnames:
                        um.append(modules[mnames.index(u)])
                    else:
                        outmess('\tModule "%s" uses nonexisting "%s" which will be ignored.\n'%(mnames[i],u))
            ret[mnames[i]] = {}
            dict_append(ret[mnames[i]],rules.buildmodule(modules[i],um))
    return ret

def dict_append(d_out,d_in):
    for (k,v) in d_in.items():
        if not d_out.has_key(k):
            d_out[k] = []
        if type(v) is types.ListType:
            d_out[k] = d_out[k] + v
        else:
            d_out[k].append(v)

def run_main(comline_list,quiet=1):
    """Run f2py as if string.join(comline_list,' ') is used as a command line.
    In case of using -h flag, return None.
    """
    reload(crackfortran)
    f2pydir=os.path.dirname(os.path.abspath(cfuncs.__file__))
    fobjhsrc = os.path.join(f2pydir,'src','fortranobject.h')
    fobjcsrc = os.path.join(f2pydir,'src','fortranobject.c')
    files,options=scaninputline(comline_list)
    postlist=callcrackfortran(files,options)
    isusedby={}
    for i in range(len(postlist)):
        if postlist[i].has_key('use'):
            for u in postlist[i]['use'].keys():
                if not isusedby.has_key(u): isusedby[u]=[]
                isusedby[u].append(postlist[i]['name'])
    for i in range(len(postlist)):
        if postlist[i]['block']=='python module' and string.find(postlist[i]['name'],'__user__')<0:
            if isusedby.has_key(postlist[i]['name']):
                if not quiet:
                    outmess('Skipping Makefile build for module "%s" which is used by %s\n'%(postlist[i]['name'],string.join(map(lambda s:'"%s"'%s,isusedby[postlist[i]['name']]),',')))
            else:
                if options['dosetup']:
                    options['setupopts'].append('-I'+os.path.dirname(fobjhsrc))
                    options['setupopts'].append(fobjcsrc)
                    buildsetup.build(postlist[i],options)
                if options['domakefile']:
                    buildmakefile.build(postlist[i],options)
    if options.has_key('signsfile'):
        if not quiet:
            outmess('Stopping. Edit the signature file and then run f2py on the signature file: ')
            outmess('%s %s\n'%(os.path.basename(sys.argv[0]),options['signsfile']))
            outmess('\tOr run GNU make to build shared module: gmake -f Makefile-<modulename>\n')
            outmess('\tOr run: python setup_<modulename>.py build\n')
        return
    for i in range(len(postlist)):
        if postlist[i]['block']!='python module':
            #errmess('All blocks must be module blocks but got %s. Exiting.\n'%(`postlist[i]['block']`))
            if not options.has_key('python module'):
                errmess('Tip: If your original code is Fortran 77 then you must use -m option.\n')
            raise TypeError,'All blocks must be module blocks but got %s'%(`postlist[i]['block']`)
    auxfuncs.debugoptions=options['debug']
    f90mod_rules.options=options
    auxfuncs.wrapfuncs=options['wrapfuncs']

    ret=buildmodules(postlist)

    if not quiet:
        outmess('Run GNU make to build shared modules: gmake -f Makefile-<modulename> [test]\n')
        outmess('Or run: python setup_<modulename>.py build\n')
    for mn in ret.keys():
        dict_append(ret[mn],{'csrc':fobjcsrc,'h':fobjhsrc})
    return ret

def filter_files(prefix,suffix,files,remove_prefix=None):
    """
    Filter files by prefix and suffic.
    """
    filtered,rest = [],[]
    match = re.compile(prefix+r'.*'+suffix+r'\Z').match
    if remove_prefix:
        ind = len(prefix)
    else:
        ind = 0
    for file in map(string.strip,files):
        if match(file): filtered.append(file[ind:])
        else: rest.append(file)
    return filtered,rest

def run_compile():
    """
    Do it all in one call!
    """
    import tempfile,os,shutil

    i = sys.argv.index('-c')
    del sys.argv[i]

    f2py_flags = filter(re.compile(r'[-][-]((no|)[-](wrap[-]functions|external[-]modroutines)|debug[-]capi)|[-]include').match,sys.argv[1:])
    sys.argv = filter(lambda a,flags=f2py_flags:a not in flags,sys.argv)
    
    flib_flags = filter(re.compile(r'[-][-](f(90)?compiler([-]exec|))=').match,sys.argv[1:])
    sys.argv = filter(lambda a,flags=flib_flags:a not in flags,sys.argv)

    modulename = 'untitled'
    if '-m' in sys.argv:
        i = sys.argv.index('-m')
        modulename = sys.argv[i+1]
        del sys.argv[i+1],sys.argv[i]

    tmp_dir = os.path.join(tempfile.mktemp())
    if not os.path.exists(tmp_dir):
        os.mkdir(tmp_dir)
    f2py_opts = ['--build-dir',tmp_dir] + f2py_flags
    sources = sys.argv[1:]
    extra_objects, sources = filter_files('','[.](o|a|so)',sources)
    include_dirs, sources = filter_files('-I','',sources,remove_prefix=1)
    library_dirs, sources = filter_files('-L','',sources,remove_prefix=1)
    libraries, sources = filter_files('-l','',sources,remove_prefix=1)
    undef_macros, sources = filter_files('-U','',sources,remove_prefix=1)
    define_macros, sources = filter_files('-D','',sources,remove_prefix=1)
    for i in range(len(define_macros)):
        name_value = string.split(define_macros[i],'=',1)
        if len(name_value)==1:
            name_value.append(None)
        if len(name_value)==2:
            define_macros[i] = tuple(name_value)
        else:
            print 'Invalid use of -D:',name_value
    from scipy_distutils.core import setup,Extension
    ext = Extension(modulename,sources,
                    include_dirs = include_dirs,
                    library_dirs = library_dirs,
                    libraries = libraries,
                    define_macros = define_macros,
                    undef_macros = undef_macros,
                    extra_objects = extra_objects,
                    f2py_options = f2py_opts,
                    )

    sys.argv = [sys.argv[0]]
    if flib_flags:
        sys.argv.extend(['build_flib']+flib_flags)
    sys.argv.extend(['build',
                     '--build-temp',tmp_dir,
                     '--build-base',tmp_dir,
                     '--build-platlib','.'])
    print libraries
    setup(ext_modules = [ext])

    if os.path.exists(tmp_dir):
        shutil.rmtree(tmp_dir)

def main():
    if '-c' in sys.argv[1:]:
        run_compile()
    else:
        run_main(sys.argv[1:],quiet=0)

#if __name__ == "__main__":
#    main()


# EOF
